from typing import Optional
from pathlib import Path

import asyncio
from functools import wraps, partial
from urllib.request import urlretrieve

from .urls import URL

def async_wrap(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)
    return run 

@async_wrap
def download_ftp(url: URL, path: Path) -> Optional[str]:
    try:
        out_path, headers = urlretrieve(str(url), str(path))
    except Exception as e:
        return str(e)
    return None



"""
Defines template strings and files for URL substitution.

See subst.py for code that performs the actual
substitution.
"""
from typing import Mapping, Sequence, Union, Tuple
from pathlib import Path

from .urls import URL

def parse_template(t : str) -> Tuple[Sequence[str], Sequence[URL]]:
    # Return a parsed form of the template string
    # as a sequence of strings, in-between which 
    # the URL-s should be inserted.
    #
    # The intended output starts and ends with a str
    # so that len(texts) == len(uris)+1
    start = '${{'
    end = '}}'
    ls = len(start)
    le = len(end)

    texts = []
    uris = []
    # Consume all instances of '${{ ... }}' template in string,
    # parsing all URL-s in-between.
    while True:
        i = t.find(start)
        if i == -1:
            break
        j = t[i+ls:].find(end)+i+ls
        if j == -1:
            raise SyntaxError(f"Missing '{end}' in '{t}'")

        texts.append( t[:i] )
        uris.append( URL(t[i+ls:j].strip()) )
        t = t[j+le:]

    texts.append(t)

    return texts, uris

class Template:
    """ Class encapsulating a string to be templated.

        Segments the input string (t) into
        self.texts and self.urls
        with len(self.texts) == len(self.urls)+1
    """
    def __init__(self, t : str):
        texts, uris = parse_template(t)
        self.texts = texts
        self.uris = uris

    def subst(self, cache : Mapping[URL, Path]) -> str:
        # substitute the template
        ans = self.texts[0]
        for u,t in zip(self.uris, self.texts[1:]):
            ans = ans + str(cache[u]) + t
        return ans

class TemplateFile(Template):
    # Class encapsulating a file to be templated.
    def __init__(self, f : Union[str, Path]):
        self.f = Path(f)
        super().__init__(self.f.read_text(encoding='utf-8'))

    def write(self, out : Union[str, Path], cache : Mapping[URL, Path]) -> None:
        # Over-write the input file with the mapped result.
        with open(out, 'w', encoding='utf-8') as f:
            f.write(self.subst(cache))



"""Substitute a template file.
"""

__author__ = "David M. Rogers"
__copyright__ = "UT-Battelle LLC"
__license__ = "BSD3"

from pathlib import Path
from typing import Optional, List, Set
import logging
_logger = logging.getLogger(__name__)

import typer

from .mirror import Mirror
from .template import TemplateFile
from .urls import URL
from . import arun

app = typer.Typer()

@app.command(help="Fetch and substitute URLs into a template.")
def subst(templates  : List[Path] = typer.Argument(..., help="File(s) to substitute."),
          results    : bool = typer.Option(False, help="Don't substitute, but list required results."),
          mirror     : Optional[Path] = typer.Option(None, help="directory holding downloaded files"),
          v     : bool = typer.Option(False, "-v", help="show info-level logs"),
          vv    : bool = typer.Option(False, "-vv", help="show debug-level logs"),
         ):
    if vv:
        logging.basicConfig(level=logging.DEBUG)
    elif v:
        logging.basicConfig(level=logging.INFO)
    if mirror is None:
        mirror = Path()

    urls : Set[URL] = set()
    outputs = {}
    for fname in templates:
        # remove last suffix
        out = fname.parent / fname.stem
        if out in outputs:
            continue
        tf = TemplateFile(fname)
        urls |= set(tf.uris)
        outputs[out] = tf

    if results:
        for url in urls:
            if url.scheme == 'result':
                assert url.s[:9] == 'result://'
                print(url.s[9:])
                #print('git' + url.s[6:])
        return 0

    M = Mirror( mirror )
    lookup = arun(M.fetch_all(urls))
    for out, tf in outputs.items():
        tf.write(out, lookup)

    return 0

if __name__ == "__main__":
    subst()



# This stand-alone file server can be used in a pinch to
# serve a directory of files.  Note it serves everything under
# the directory you are running from!
#
# It supports HEAD queries and partial file downloads
# as used by aurl's parallel download methods.

import os, sys
from typing import Union, Dict
from pathlib import Path, PurePath

from dataclasses import dataclass
@dataclass
class FileStat:
    size: int
    atime: int
    mtime: int
    children: Union[bool, Dict[str,"FileStat"]]

def stat_dir(path: Path, max_depth=0) -> Dict[str, FileStat]:
    # Caution! the path is not checked to ensure
    # it is safe to serve. (caller should do this)
    ans = {}
    for p in path.iterdir():
        st = p.stat()
        ans[p.name] = FileStat( size = int(st.st_size),
                                atime = int(st.st_atime),
                                mtime = int(st.st_mtime),
                                children = False,
                              )
        if p.is_dir():
            if max_depth > 0:
                ans[p.name].children = stat_dir(p, max_depth-1)
            else:
                ans[p.name].children = True
    return ans

try: # fastapi is optional
    from fastapi import FastAPI, HTTPException, Response # type: ignore[import-not-found]
    from fastapi.responses import FileResponse # type: ignore[import-not-found]
    app = FastAPI()

    try: # improved logging is optional
        from certified.formatter import log_request # type: ignore[import-not-found]
        app.middleware("http")(log_request)
    except ImportError:
        pass

except ImportError: # These stubs will allow this module to load, but not work.
    class App(): # dummy app
        def get(self, *args, **kws):
            # empty decorator
            return lambda fn: fn
        def head(self, *args, **kws):
            # empty decorator
            return lambda fn: fn
    app = App() # type: ignore[assignment]
    class HTTPException(Exception): # type: ignore[assignment, no-redef]
        def __init__(self, status_code, detail):
            super().__init__(detail)

file_root = Path().resolve()

def safe_path(base: Path, fname: Union[Path,str]) -> Path:
    """ Take an unsafe, user-provided fname, validate it,
        and place it relative to the base path.

        Throw an exception if the path contains ".." or is absolute.
    """

    rel = PurePath(fname)
    if rel.is_absolute() or ".." in rel.parts:
        raise HTTPException(status_code=403, detail="invalid path")

    if not base.is_dir():
        raise HTTPException(status_code=404, detail="base dir missing")

    # FIXME: check whether this path traverses a symlink
    # (https://stackoverflow.com/questions/41460434/getting-the-target-of-a-symbolic-link-with-pathlib)
    return base / rel

@app.get("/{filename:path}")
async def get_file(filename: str, max_depth: int = 0):
    """
    Serves a file from the working directory if it exists.
    """
    file_path = safe_path(file_root, filename)
    if file_path.is_file():
        return FileResponse(file_path, filename=file_path.name)
    elif file_path.is_dir():
        max_depth = min(max_depth, 3) # truncate to at most 3
        return stat_dir(file_path, max_depth)
    else:
        raise HTTPException(status_code=404, detail="File not found")

@app.head("/{filename:path}")
async def head_file(filename: str):
    """
    Handles HEAD requests for files in the working directory.
    Returns headers without the file body.
    """
    p = safe_path(file_root, filename)
    try:
        stat = p.stat()
    except FileNotFoundError:
        raise HTTPException(status_code=404, detail="File not found")

    hdr = {
        "content-length": str(stat.st_size),
        "content-type": "application/octet-stream",
        "accept-ranges": "bytes",
        "content-disposition": f"attachment; filename={p.name}",
    }
    return Response(status_code=200, headers=hdr)



from typing import Optional, Union
from pathlib import Path
import os, logging

from urllib.parse import urlparse, parse_qs, quote, unquote

class URL:
    """Parsed URL object.

    Fully parse a string into URL components
    and validate against valid URL formats.
   
    Attributes:

    * scheme   : str
    * netloc   : str
    * path     : str
    * query    : {key:val}
    * fragment : str

    """
    def __init__(self, s1 : Union[str, 'URL'], validate=True):
        if isinstance(s1, str):
            s = s1
        else:
            s = s.s # type: ignore[attr-defined]
        ans = urlparse(s, scheme='', allow_fragments=True)
        
        # store metadata
        if ans.query != "":
            self.meta = f"?{ans.query}"
        else:
            self.meta = ""
        if ans.fragment != "":
            self.meta = f"{self.meta}#{ans.fragment}"

        self.scheme = unquote(ans.scheme)
        self.netloc = unquote(ans.netloc)
        assert "%2F" not in ans.path, "Invalid path."
        self.path = unquote(ans.path)
        if self.scheme != "file": # remove leading '/' in paths
            if self.path.startswith("/"):
                self.path = self.path[1:]
        if len(ans.query) > 0:
            self.query = parse_qs(ans.query,
                              keep_blank_values=True,
                              strict_parsing=True,
                              errors="strict")
        else:
            self.query = {}
        self.fragment = unquote(ans.fragment)
        self.s = ans.geturl()
        if not validate:
            return
        try:
            self.validate()
        except AssertionError as e:
            raise AssertionError(f"Invalid URL format: {self.s} -- {e}")

    def with_scheme(self, scheme):
        return urlparse(self.s, scheme="", allow_fragments=True) \
                  . _replace(scheme=scheme) \
                  . geturl()

    def __repr__(self):
        return f"URL('{self.s}')"
    def __str__(self):
        return self.s
    def __hash__(self):
        return hash(repr(self))
    def __eq__(a, b):
        return repr(a) == repr(b)
    def fullpath(self):
        if self.scheme == "file":
            return self.path
        s = self.netloc
        if len(self.path) > 0:
            s += '/' + self.path
        return s
    def validate(url):
        absent = lambda x: len(getattr(url, x)) == 0
        # netloc, path, query, fragment
        if url.scheme in ["file", "git", "git+file", "git+http",
                          "git+https", "git+ssh", "ftp"]:
            assert absent("query") and absent("fragment")
        elif url.scheme in ["result"]:
            assert absent("query")
        elif url.scheme == "https" or url.scheme == "http":
            assert absent("fragment")
        else:
            raise AssertionError(f"Unknown URL scheme: {url.scheme}")




__author__ = "David M. Rogers"
__copyright__ = "UT-Battelle LLC"
__license__ = "BSD3"

from typing import List, Optional, Dict, Any
from pathlib import Path
import asyncio
import logging
import sys
_logger = logging.getLogger(__name__)

import typer
import json

from .mirror import Mirror
from .urls import URL
from . import arun

app = typer.Typer()

async def get_list(url: str, M: Mirror, max_depth: int = 3,
                   ignore_hidden: bool = True) -> List[URL]:
    loc = await M.fetch(URL(f"{url}?max_depth={max_depth}"))
    if loc is None:
        print(f"Unable to download file listing for {url}")
        sys.exit(1)
    with loc.open() as f:
        tree = json.load(f)

    async def add_tree(path: str, t: Dict[str,Any], urls: List[URL]):
        for name, entry in t.items():
            if name.startswith(".") and ignore_hidden:
                continue
            rel = f"{path}/{name}"
            tree = entry.get('children', False)
            if tree:
                if tree is True:
                    loc = await M.fetch(URL(f"{rel}?max_depth={max_depth}"))
                    if loc is None:
                        print(f"Unable to download file listing for {rel}")
                        continue
                    with loc.open() as f:
                        tree = json.load(f)
                await add_tree(rel, tree, urls)
                continue
            urls.append(URL(rel))

    urls: List[URL] = []
    await add_tree(url, tree, urls)
    return urls

@app.command(help="Get a directory structure served by aurl.serve.")
def get_dir(url    : str = typer.Argument(..., help="directory tree root"),
            mirror : Optional[Path] = typer.Option(None, help="directory holding downloaded files"),
            v    : bool = typer.Option(False, "-v", help="show info-level logs"),
            vv   : bool = typer.Option(False, "-vv", help="show debug-level logs")):
    if vv:
        logging.basicConfig(level=logging.DEBUG)
    elif v:
        logging.basicConfig(level=logging.INFO)
    if mirror is None:
        mirror = Path()

    M = Mirror( mirror )

    urls = arun( get_list(url, M) )
    paths = arun( M.fetch_all(urls) )
    print(json.dumps(dict((k.s, str(v)) for k, v in paths.items()), indent=4))
    sys.exit(0)

if __name__=="__main__":
    app()



from typing import Optional, Dict, Tuple, Union
import logging
_logger = logging.getLogger(__name__)
import time
from urllib.parse import urlsplit, urlunsplit

from pathlib import Path
import asyncio

import aiohttp
import aiofiles

from .exceptions import DownloadException
from .urls import URL
from .search import which, lookup_local
from .runcmd import runcmd
from .aftp import download_ftp

Pstr = Union[str, Path]

class UnsupportedOperation(Exception):
    pass

async def download_part(session: aiohttp.ClientSession, url: str, dest: Pstr,
                        start: int, end: int, chunk_size: int):
    """ Download part of this URL using a Range header request
        between start and end (slice-like, 0-indexed, non-inclusive).

        Write the result to the destination file at the starting offset.

        Raises UnsupportedOperation if the download should be re-tried in serial.

        Raises DownloadException on other responses.
    """
    assert start >= 0 and end > start, "Invalid range"
    headers = {"Range": f"bytes={start}-{end-1}"}
    async with session.get(url, allow_redirects=True, headers=headers) as response:
        if response.status == 206:  # Partial Content
            async with aiofiles.open(dest, mode="r+b") as f:
                await f.seek(start)
                async for chunk in response.content.iter_chunked(chunk_size):
                    await f.write(chunk)
        elif response.status in [200, 501]: # ignored / not implemented
            _logger.info("%s: GET with Range failed with %d", url, response.status)
            raise UnsupportedOperation()
        else:
            raise DownloadException("Download error on %s (%d-%d): received status %d"%
                                    (url, start, end, response.status))

async def download_full(session: aiohttp.ClientSession, url: str, dest: Pstr,
                        chunk_size: int):
    """ Download the URL contents to file.

        Raises DownloadException on error.
    """
    async with session.get(url, allow_redirects=True) as response:
        if response.status != 200:
            raise DownloadException("Download error on %s: received status %d"%
                                    (url, response.status))
        async with aiofiles.open(dest, mode="wb") as f:
            async for chunk in response.content.iter_chunked(chunk_size):
                await f.write(chunk)
            return await f.tell()

# try 1024**2 or 8192...
async def download_url(outfile: Pstr,
                       url1: Union[str, URL],
                       chunk_size: int = 1024**2,
                       max_connections: int = 4) -> int:
    """ Download the url to the given output file.

        Raises a DownloadException on error.

        Returns the downloaded file size (in bytes) on success.
    """
    assert chunk_size > 0 and max_connections > 0
    dest = Path(outfile)
    dest.parent.mkdir(exist_ok=True, parents=True)

    # Rewrite the URL so that the scheme and netloc appear in the base.
    (scheme, netloc, path, query, fragment) = urlsplit(str(url1))
    base = urlunsplit((scheme, netloc,"","",""))
    url  = urlunsplit(("","",path,query,fragment))

    try:
        from certified import Certified # type: ignore[import-not-found]
        mk_session = Certified().ClientSession
    except ImportError:
        mk_session = aiohttp.ClientSession # type: ignore[assignment]

    file_size: Optional[int] = None
    async with mk_session(base) as session:
        async with session.head(url, allow_redirects=True) as response:
            if response.status == 200:
                if 'Content-Length' in response.headers:
                    file_size = int(response.headers.get('Content-Length', 0))

        if file_size is None:
            async with session.get(url, allow_redirects=True) as response:
                if response.status != 200:
                    raise DownloadException("%s: Error getting size (%d): %s"%(
                                            url1, response.status, response.text()))
                if 'Content-Length' in response.headers:
                    file_size = int(response.headers.get('Content-Length', 0))
                else: # just download the file
                    async with aiofiles.open(dest, mode="wb") as f:
                        async for chunk in response.content.iter_chunked(chunk_size):
                            await f.write(chunk)
                        return await f.tell()

        if file_size == 0:
            async with aiofiles.open(dest, mode="wb") as f:
                pass
            return 0
        # Create an empty file with the total size
        async with aiofiles.open(dest, mode="wb") as f:
            await f.seek(file_size - 1)
            await f.write(b'\0')

        chunks = (file_size+chunk_size-1)//chunk_size
        connections = min(chunks, max_connections)

        # note we always have chunks >= connections
        data_per_task = ( chunks // connections ) * chunk_size

        tasks = []
        for i in range(connections):
            start =   i * data_per_task
            end = start + data_per_task
            if i == connections-1:
                end = file_size  # Ensure the last part goes to the end
            task = asyncio.create_task(download_part(session, url, dest,
                                                     start, end, chunk_size))
            tasks.append(task)

        try:
            await asyncio.gather(*tasks)
        except UnsupportedOperation:
            # run the complicated cleanup from partially complete gather.
            [t.cancel() for t in tasks if not t.done()]
            for t in tasks:
                if not t.done():
                    try:
                        await t
                    except (asyncio.CancelledError, UnsupportedOperation):
                        pass
            return await download_full(session, url, dest, chunk_size)
    return file_size

async def lookup_or_fetch(url : URL, hostname : str, base : Path) -> Path:
    # Resolve URL and download.
    #
    # Base is the location to store the final result
    # in the event a download is needed.
    # This function is responsible for running mkdir/etc.
    # to get to base.
    #
    # Returns a local path if the resource can be retrieved
    # successfully, or None if the resource cannot be downloaded.
    #
    # Handles the following URL types:
    #    - (http|https)://* - download with aiohttp
    #    - git://* - run git clone
    #    - git+(http|https|ssh)://* - run git clone
    #    - file://* TODO - check multiple filesystems
    #    - result://* TODO - escalate to higher-level servers
    #
    #
    # May throw a DownloadException
    #
    ans = await lookup_local(url, hostname)
    if ans is not None:
        return ans
    _logger.info("Attempting to download %s", url)
    if url.scheme == "ftp":
        err = await download_ftp(url, base)
        if err:
            raise DownloadException(err)
        return base
    elif url.scheme == "http" or url.scheme == "https":
        t0 = time.time()
        sz = await download_url(base, url.s)
        dt = time.time() - t0
        _logger.info("%s: %d bytes at %f Mbps", url, sz, sz*8/1024**2/dt)
        return base
    elif url.scheme.startswith("git"):
        base.parent.mkdir(exist_ok=True, parents=True)
        gurl = url.s
        if gurl.startswith("git+"):
            gurl = gurl[4:]
        if '@' in url.s:
            gurl, commit = gurl.split('@', 1)
            ret, out, err = await runcmd("git", "clone", "--branch",
                                         commit, gurl, str(base))
        else:
            ret, out, err = await runcmd("git", "clone", gurl, str(base))
        if ret != 0:
            raise DownloadException(err)
        return base
    elif url.scheme == "file":
        if url.netloc == hostname or len(url.netloc) == 0:
            #return '/'+url.path
            ans = Path(url.path)
            if not ans.exists():
                raise DownloadException(f"Path does not exist: {url.s}")
            return ans
        # try to fetch remote file://?
        raise DownloadException(f"Unreachable {url.s}")
    elif url.scheme == "result":
        # These must be reached through a Mirror.
        raise DownloadException(f"Path does not exist: {url.s}")

    _logger.error("Unknown scheme, can't fetch: %s", url)
    raise DownloadException("Unknown scheme, can't fetch: %s"%url)



from typing import Optional
from functools import cache
from pathlib import Path
import os
import logging
_logger = logging.getLogger(__name__)

from .exceptions import DownloadException
from .urls import URL
from .runcmd import runcmd

# https://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
@cache
def which(program : str) -> Optional[Path]:
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return Path(program)
    else:
        for path in os.environ.get("PATH", "").split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return Path(exe_file)

    return None

async def lookup_local(url : URL, hostname : str) -> Optional[Path]:
    # Note: Several of the lookups performed print
    #       error messages to stderr.  We do not
    #       capture these, but allow them to pass-through.
    #
    # Handles the following cases
    #    - file://{hostname}/*
    #
    # Returns:
    #  * Path on success
    #  * None if further lookup is needed
    # 
    # May throw a DownloadException if lookup is impossible
    #
    fail = (False, None)
    if url.scheme == "file":
        if url.netloc == hostname or len(url.netloc) == 0:
            p = Path(url.path)
            if not p.exists():
                raise DownloadException(f"{url.s} does not exist locally")
            return p
        return None

    return None



from typing import Optional, Union, Dict
from collections.abc import Iterable
from pathlib import Path
import socket
import logging
_logger = logging.getLogger(__name__)

from .exceptions import DownloadException
from .urls import URL
from .fetch import lookup_or_fetch
from .taskmgr import ResourceQueue, ResourceContext, TaskMgr

def gethostname():
    #fqdn = socket.getfqdn(socket.gethostname())
    return socket.gethostname()

class Mirror:
    """Manage a local cache of data located at path `base`.
    
    the layout is::

        `base`/
           encoded1/
               files-inside-encoded1
           encoded2/ (mirrored remote subtree)
               files-inside-encoded1
           encoded3 (mirrored remote file)
           ....
    
    Encode/decode work as follows:

    >>> from pathlib import Path
    >>> from aurl.mirror import Mirror
    >>> from aurl.urls import URL
    >>> base = Path('/cache')
    >>> C = Mirror( base )
    >>> src = URL('https://www.example.com/index.html')
    >>> str( C.encode(src) )
    '/cache/https/www.example.com/index.html'

    >>> C.decode(base / 'https/www.example.com/index.html')
    URL('https://www.example.com/index.html')

    >>> src = URL('git://spool/frobnitzem/dwork/README.md#v1')
    >>> str( C.encode(src) )
    '/cache/git/spool#v1/frobnitzem/dwork/README.md'

    >>> C.decode(base / 'git/spool#v1/frobnitzem/dwork/README.md')
    URL('git://spool/frobnitzem/dwork/README.md#v1')

    >>> src = URL('http://nevada/user?tango=alpha')
    >>> str( C.encode(src) )
    '/cache/http/nevada?tango=alpha/user'

    >>> C.decode(base / 'http/nevada?tango=alpha/user')
    URL('http://nevada/user?tango=alpha')
    """
    def __init__(self, base : Union[str, Path], nparallel : int = 10):
        self.hostname = gethostname()
        self.base = Path(base).resolve()
        assert self.base.is_dir()

        self.cq = ResourceQueue(list(range(nparallel)))

        #self.db = {} #: Mapping from url to local path
        ## scan for initial database contents
        #for x in self.base.iterdir():
        #    if not x.is_dir(): continue
        #    url = self.decode(x)
        #    self.db[url] = x

    def encode(self, url : URL) -> Path:
        # write the path where the given URL would be stored
        ans = self.base / url.scheme / (url.netloc+url.meta)

        path = url.path
        # remove leading '/'
        if len(path) > 0 and path[0] == '/':
            path = path[1:]
        if len(path) > 0:
            ans = ans / path

        return ans

    def decode(self, path : Path) -> Optional[URL]:
        # write the URL which the path links to
        try:
            suffix = path.relative_to(self.base)
        except ValueError:
            return None

        p = suffix.parts
        if len(p) < 2:
            return None
        scheme = p[0]
        loc = URL(p[1], False)
        try:
            ans = URL(f"{scheme}://{loc.path}/" + '/'.join(p[2:]) + loc.meta)
        except: # validation error
            return None
        return ans

    async def fetch(self, url : URL) -> Optional[Path]:
        """Handles url downloads.

        Args:
           url: the resource to lookup or fetch.

        Returns:
           A Path pointing at the local, cached version of the
           resource (or None if lookup & fetch was unsuccessful).
        """
        if not isinstance(url, URL):
            _logger.error("get received a non-URL input")
            return None

        out = self.encode(url)
        if out.exists():
            return out

        _logger.info("No local copy of %s exists, attempting fetch.", url)
        async with ResourceContext(self.cq) as r:
            return await lookup_or_fetch(url, self.hostname, out)

    async def fetch_all(self, urls : Iterable[URL]) -> Dict[URL, Path]:
        """ Fetch all urls from the given mirror.
            Returns a mapping from url to the path where it can
            be accessed locally.

            raises DownloadException on error.
        """
        location : Dict[URL, Path] = {}
        errors = []
        with TaskMgr() as T:
            for url in set(urls):
                T.start(self.fetch(url), url)
            for t, url in T:
                try:
                    location[url] = await t
                except DownloadException as e:
                    errors.append(str(url)+": "+str(e))
        if len(errors) > 0:
            raise DownloadException("Download errors:\n  - "
                                    + "\n  - ".join(errors))

        return location

    def to_url(self, fname : Path) -> str:
        """Returns a URL representation of a local path.
        """
        return f"file://{self.hostname}{fname}"

#   def new_stem(self):
#       # create a new subdirectory to hold a subtree of
#       # data from a single source
#       return Path( mkdtemp(dir=self.path) )



from typing import Union, Tuple, Optional
import logging
_logger = logging.getLogger(__name__)

import asyncio
from pathlib import Path
from time import time as timestamp

async def runcmd(prog : Union[Path,str], *args : str,
                 cwd : Union[Path,str,None] = None,
                 expect_ok : Optional[bool] = True) -> Tuple[int,str,str]:
    """Run the given command inside an asyncio subprocess.
       
       Returns (return code : int, stdout : str, stderr : str)
    """
    pipe = asyncio.subprocess.PIPE
    proc = await asyncio.create_subprocess_exec(
                    str(prog), *args, cwd=cwd,
                    stdout=pipe, stderr=pipe)
    stdout, stderr = await proc.communicate()
    # note stdout/stderr are binary

    out = stdout.decode('utf-8')
    err = stderr.decode('utf-8')
    if len(stdout) > 0:
        _logger.debug('%s stdout: %s', prog, out)
    if len(stderr) > 0:
        _logger.info('%s stderr: %s', prog, err)

    ret = -1
    if proc.returncode is not None:
        ret = proc.returncode
    if expect_ok != (proc.returncode == 0):
        _logger.error('%s returned %d', prog, ret)
    if expect_ok is None:
        _logger.info('%s returned %d', prog, ret)

    return ret, out, err



from typing import Optional, BinaryIO
import os

import asyncio
from typing import Awaitable, TypeVar
from typing import Generic, Sequence

from pathlib import Path

class TaskMgr:
    """Helper class for storing a list of tasks.

    Stores a list of tasks with optional additional info
    in its task-list.  On exit, all pending tasks
    are cancelled if they are not yet complete.

    The TaskMgr is iterable.  Note, however, that
    each loop will go through the tasks sequentially
    without replacement.  This means (loop..break) (loop)
    will not see the same set of tasks.

    Usage::
    
        import asyncio
        from aurl.taskmgr import TaskMgr
        with TaskMgr() as T:
           T.start(asyncio.sleep(1), 's1')
           T.start(asyncio.sleep(2), 's2')
           for t, name in T:
               await t
               print(f"{name} completed")
               break

    """
    def __init__(self):
        self.tasks = []
        #self.task_info = {}
        self.cur = 0 # next task to await completion (index into self.tasks)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        # (exc_type, exc, traceback) are *sys.exc_info()
        # or else None (on normal exit)
        for t in self.tasks: #[max(self.cur-1,0):]:
            if not t[0].done():
                t[0].cancel()
        return False # continue to raise the exception

    def start(self, c : Awaitable, *info):
        """Start the coroutine c as a task.
        
        Also stores its info for later retrieval.
        """
        #t = asyncio.create_task(c)
        t = asyncio.ensure_future(c)
        self.tasks.append((t,) + tuple(info))
        #self.task_info[t] = info

    def __iter__(self):
        return self

    def __next__(self):
        if self.cur == len(self.tasks):
            raise StopIteration
        tinfo = self.tasks[self.cur]
        self.cur += 1
        return tinfo

V = TypeVar('V')
class ResourceQueue(asyncio.Queue, Generic[V]):
    """Container for `n` parallel-accessible resources.

       Example:

           # start 3 client connections
           self.cq = ResourceQueue([ APIClient(config) for i in range(3) ])

           # Send a ping over the client connection.
           # Can be called concurrently.
           async def ping():
             async with ResourceContext(self.cq) as r:
                return await r.call("ping")
    """
    def __init__(self, resources : Sequence[V]):
        asyncio.Queue.__init__(self, len(resources))
        for r in resources:
            self.put_nowait( r )

class ResourceContext(Generic[V]):
    def __init__(self, cq : ResourceQueue[V]):
        self.cq = cq
    async def __aenter__(self) -> V:
        self.r = await self.cq.get()
        return self.r
    async def __aexit__(self, exc_type, exc, traceback):
        # (exc_type, exc, traceback) are *sys.exc_info()
        # or else None (on normal exit)
        await self.cq.put(self.r)
        return False # continue to raise the exception



from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version('aurl-aurl')
except PackageNotFoundError:
    __version__ = 'dev'

import asyncio

from .exceptions import DownloadException
from .urls import URL
from .template import Template, TemplateFile
from .mirror import Mirror

def arun(f):
    loop = asyncio.get_event_loop()
    ans = loop.run_until_complete(f)
    #loop.close()
    return ans




"""Get a list of URLs.
"""

__author__ = "David M. Rogers"
__copyright__ = "UT-Battelle LLC"
__license__ = "BSD3"

from typing import List, Optional
from pathlib import Path
import asyncio
import logging
_logger = logging.getLogger(__name__)

import typer
import json

from .mirror import Mirror
from .urls import URL
from . import arun

app = typer.Typer()

@app.command(help="Download a list of URLs.")
def get(urls   : List[str] = typer.Argument(..., help="urls to download"),
        mirror : Optional[Path] = typer.Option(None, help="directory holding downloaded files"),
          v    : bool = typer.Option(False, "-v", help="show info-level logs"),
          vv   : bool = typer.Option(False, "-vv", help="show debug-level logs")):
    if vv:
        logging.basicConfig(level=logging.DEBUG)
    elif v:
        logging.basicConfig(level=logging.INFO)
    if mirror is None:
        mirror = Path()

    M = Mirror( mirror )
    urls1 = [URL(u) for u in urls]
    paths = arun(M.fetch_all(urls1))
    print(json.dumps(dict((k.s, str(v)) for k, v in paths.items()), indent=4))

if __name__=="__main__":
    get()
